import os

import hvac
from hvac.exceptions import VaultError

from mind_castle.secret_store_base import SecretStoreBase
from mind_castle.exceptions import GetSecretException, PutSecretException


class HashiCorpVaultSecretStore(SecretStoreBase):
    """
    Uses HashiCorp Vault to store secrets.
    """

    store_type = "hashicorpvault"
    required_config = [["MIND_CASTLE_VAULT_HOST", "MIND_CASTLE_VAULT_TOKEN"]]

    def __init__(self):
        self.client = hvac.Client(
            url=os.environ.get("MIND_CASTLE_VAULT_HOST"),
            token=os.environ.get("MIND_CASTLE_VAULT_TOKEN"),
        )
        self.mount_point = "cubbyhole/"

    def get_secret(self, secret_details: dict) -> str:
        key = secret_details.get("key")
        try:
            response = self.client.secrets.kv.v1.read_secret(
                path=key, mount_point=self.mount_point
            )
        except VaultError as e:
            raise GetSecretException(e, key)

        return response["data"]["data"].get("secret_value")

    def put_secret(self, value: str) -> dict:
        key = self.get_secret_key()
        try:
            self.client.secrets.kv.v1.create_or_update_secret(
                path=key, secret=dict(secret_value=value), mount_point=self.mount_point
            )  # Value has to be a dict so just make a 'secret_value' key
        except VaultError as e:
            raise PutSecretException(e, key)

        return {"mind_castle_secret_type": self.store_type, "key": key}



import os

import boto3
import botocore

from mind_castle.secret_store_base import SecretStoreBase
from mind_castle.exceptions import GetSecretException, PutSecretException


class AWSSecretsManagerSecretStore(SecretStoreBase):
    """
    Uses AWS Secrets Manager to store secrets.
    This assumes you have credentials in the environment or in the default AWS credentials file.
    """

    store_type = "awssecretsmanager"
    required_config = [
        [
            "MIND_CASTLE_AWS_REGION",
            "MIND_CASTLE_AWS_ACCESS_KEY_ID",
            "MIND_CASTLE_AWS_SECRET_ACCESS_KEY",
        ],
        ["MIND_CASTLE_AWS_REGION", "MIND_CASTLE_AWS_USE_ENV_AUTH"],
    ]
    optional_config = ["MIND_CASTLE_AWS_SECRET_KEY_PREFIX"]

    def __init__(self):
        if all([os.environ.get(req) for req in self.required_config[0]]):
            # Configure with secret key
            self.client = boto3.client(
                "secretsmanager",
                region_name=os.environ["MIND_CASTLE_AWS_REGION"],
                aws_access_key_id=os.environ["MIND_CASTLE_AWS_ACCESS_KEY_ID"],
                aws_secret_access_key=os.environ["MIND_CASTLE_AWS_SECRET_ACCESS_KEY"],
            )
        else:
            # Assume the environment is configured with the correct credentials
            self.client = boto3.client(
                "secretsmanager", region_name=os.environ["MIND_CASTLE_AWS_REGION"]
            )

        self.secret_key_prefix = os.environ.get("MIND_CASTLE_AWS_SECRET_KEY_PREFIX", "")

    def get_secret(self, secret_details: dict) -> str:
        key = secret_details.get("key")
        try:
            response = self.client.get_secret_value(SecretId=secret_details.get("key"))
        except botocore.exceptions.ClientError as e:
            raise GetSecretException(e, key)

        return response.get("SecretString")

    def put_secret(self, value: str) -> dict:
        key = self.secret_key_prefix + self.get_secret_key()
        # Response is always a success. This throws an exception if it fails
        try:
            response = self.client.create_secret(
                Name=key,
                ClientRequestToken=key,
                SecretString=value,
                # Other details we could add:
                # Description='string',
                # KmsKeyId='string',
                # SecretBinary=b'bytes',
                # Tags=[
                #     {
                #         'Key': 'string',
                #         'Value': 'string'
                #     },
                # ],
                # AddReplicaRegions=[
                #     {
                #         'Region': 'string',
                #         'KmsKeyId': 'string'
                #     },
                # ],
                # ForceOverwriteReplicaSecret=True|False
            )
        except botocore.exceptions.ClientError as e:
            raise PutSecretException(e, key)

        # ARN is not needed but returned just in case. we could also add tags etc to this response
        return {
            "mind_castle_secret_type": self.store_type,
            "key": key,
            "arn": response.get("ARN"),
        }



from unittest.mock import ANY

from mind_castle.stores.vault import HashiCorpVaultSecretStore


def test_put_secret(mocker):
    secret_store = HashiCorpVaultSecretStore()
    mocker.patch.object(
        secret_store.client.secrets.kv.v1, "create_or_update_secret", return_value={}
    )
    response = secret_store.put_secret("some_secret_value")

    assert secret_store.client.secrets.kv.v1.create_or_update_secret.called_once_with(
        secret=dict(path=ANY, secret_value="some_secret_value")
    )
    assert response["mind_castle_secret_type"] == secret_store.store_type
    assert response["key"] is not None


def test_get_secret(mocker):
    mock_vault_response = {
        "data": {
            "data": {
                "secret_value": "some_secret_value",
            }
        }
    }

    secret_store = HashiCorpVaultSecretStore()
    mocker.patch.object(
        secret_store.client.secrets.kv.v1,
        "read_secret",
        return_value=mock_vault_response,
    )
    response = secret_store.get_secret(
        {"mind_castle_secret_type": "hashicorpvault", "key": "some_secret_key2"}
    )

    assert secret_store.client.secrets.kv.v1.read_secret.called_once_with(
        path="some_secret_key2"
    )
    assert response == "some_secret_value"



import json
import logging

from sqlalchemy import types

from mind_castle.stores import get_secret, put_secret
from mind_castle.exceptions import PutSecretException

logger = logging.getLogger(__name__)
logger.level = logging.DEBUG


class SecretData(types.TypeDecorator):
    """A sqlalchemy field type that stores data in a secret store."""

    impl = types.JSON  # The base data type for this field
    cache_ok = True

    def __init__(self, store_type: str, secret_key_prefix: str = "", *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.secret_type = store_type
        self.secret_key_prefix = secret_key_prefix

    def process_bind_param(self, value, dialect):
        """Convert a plaintext value into a secret and return it to be stored."""
        # Don't waste time storing empty values
        if value is None or value == "" or value == {} or value == []:
            return value

        # Make a string out of whatever object we got
        stringValue = json.dumps(value)

        try:
            secret_params = put_secret(stringValue, self.secret_type)
            logger.debug(f"Stored secret with: {secret_params}")
        except PutSecretException as e:
            logger.error(
                f"Failed to store secret in {self.secret_type}. Falling back to plaintext."
            )
            logger.error(e)

            secret_params = put_secret(stringValue, "plaintext")
            logger.warning(f"Stored secret IN PLAINTEXT with: {secret_params}")

        return secret_params

    def process_result_value(self, value, dialect):
        """Convert a secret from the DB into the original value."""

        if value is None:
            return None

        # Get a dict of the secret details
        secret_details = json.loads(value) if isinstance(value, str) else value

        # This might be an old plaintext value
        if secret_details.get("mind_castle_secret_type") is None:
            logger.debug(
                f"No secret type found in '{secret_details}', must be a plaintext value."
            )
            return secret_details

        logger.debug(f"Restoring {secret_details} from a secret.")
        return json.loads(get_secret(secret_details))



from mind_castle.stores.plaintext import PlaintextSecretStore


def test_put_secret():
    secret_store = PlaintextSecretStore()
    response = secret_store.put_secret("some_secret_value")

    # Just read the memory from this implementation
    assert response["key"] == "some_secret_value"
    assert response["mind_castle_secret_type"] == secret_store.store_type


def test_get_secret():
    secret_store = PlaintextSecretStore()
    response = secret_store.put_secret("some_secret_value")

    assert (
        secret_store.get_secret(
            {"mind_castle_secret_type": "plaintext", "key": response["key"]}
        )
        == "some_secret_value"
    )



import traceback


class GetSecretException(Exception):
    """Raised when a secret store operation fails."""

    def __init__(self, inner_exception: Exception, secret_key: str):
        message = f"Failed to get secret with key: {secret_key}. Secret store exception:\n{''.join(traceback.format_exception(inner_exception))}"
        super().__init__(message)


class PutSecretException(Exception):
    """Raised when a secret store operation fails."""

    def __init__(self, inner_exception: Exception):
        message = f"Failed to put secret. Secret store exception:\n{''.join(traceback.format_exception(inner_exception))}"
        super().__init__(message)



import logging

from mind_castle.secret_store_base import SecretStoreBase

logger = logging.getLogger(__name__)


# DO NOT USE THIS STORE IN PRODUCTION
# It will store your secrets IN PLAIN TEXT in the DB


class PlaintextSecretStore(SecretStoreBase):
    """
    A nop secret store that just stores the given plaintext in the DB.
    Use only for testing or development.
    """

    store_type = "plaintext"

    def get_secret(self, secret_details: dict) -> str:
        return secret_details.get("key")

    def put_secret(self, value: str) -> dict:
        return {"mind_castle_secret_type": self.store_type, "key": value}
